/* stp_user_int.c contains the functions for the STP programmatic
 * interface that user's should interact with
 * 
 *
 * Completed 01/29/04 - N. Mullaney
 */


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "sac.h"
#include "stp_user_int.h"


static char **defaults = NULL;
static int numdef = 0;
static char server[MAXLINE] = "*";

// flags to keep track of what type of data we have
static int TRIG_WF = 0;
static int GOT_DIR = 0;
static int WIN_WF = 0;
static int EVENT = 0;
static int ALTLOC = 0;
static int PHASE = 0;
static int STATION = 0;
static int DO_LONG = 0;
static int DO_SHORT = 0;


/* Housekeeping functions */

/* set a list of commands (separated by a ;) to be issued when the 
 * connection to the wave server is opened
 */
void stp_set_defaults(char *default_list) {
  int i, j, k;
  char **d;
  int words;
  
  // count the number of words (i.e. # of ; + 1)
  words = 0;
  for(i = 0; i< strlen(default_list); i++) {
    if (default_list[i]==';') {
      words++;
    }
  }
  words++;

  // allocate space for the word array
  d = realloc(defaults, words);
  if (d == NULL) {
    fprintf(stderr, "Could not allocate space for defaults\n");
    return;
  }
  defaults = d;
  d = NULL;

  // allocate space for each word
  for (i = 0; i< words; i++) {
    defaults[i] = (char *) malloc(MAXLINE);
    if (defaults[i] == NULL) {
      for (j = 0; j < i; j++) {
	free(defaults[j]);
      }
      return;
    }
  }
  
  // move words into defaults
  j = 0; k = 0;
  for (i = 0; i < strlen(default_list); i++) {
    if (default_list[i] == ';') {
      defaults[j][k] = '\0';
      k = 0;
      j++;
    }
    else {
      defaults[j][k] = default_list[i];
      k++;
    }
  }

  numdef = words;
 
  return;
}

/* sends the current list of defaults to the waveserver 
 */
void send_defaults () {
  int i, n;
  char temp[MAXLINE];
  strcpy(temp, "");

  if (defaults == NULL) 
    return;
 
  for (i = 0; i < numdef; i++) {
    send_stp_request(defaults[i]);
    // read any response the server might have
    while(n = stp_string_read(temp) != 0);
    if(n == -1) {
      fprintf(stderr, "Error writing defaults: %s\n", temp);
    }
  }
  
  return;
}

/* frees the memory of the defaults */
void free_defaults () {
  int i;
  for (i = 0; i < numdef; i++) {
    free(defaults[i]);
  }
  free(defaults);
  return;
}



/* sets a preferred server to use when asking
 * for a connection
 */
void stp_set_server(char *server_name) {

  strcpy(server_name, server);
  return;
}

/****************************************************************************/
/*                         WAVEFORMS                                         */
/*****************************************************************************/

/* send a request to the stpserver for waveforms
 */
int stp_req_waveform (char *request_command) {
  int n;
  
  // make sure we have no data waiting
  // to prevent nested requests
  if (WIN_WF==1 || TRIG_WF==1 || EVENT==1 || ALTLOC==1 || PHASE==1 || STATION==1) {
    fprintf(stderr, "Waveform request failed: another request is in progress.\n");
    return -1;
  }
  
  //open the waveserver
  open_waveserver(server);
  
  // send defaults 
  send_defaults();
  free_defaults();
  
  // check that we are sending an appropriate command
  if (strncmp(request_command, "trig", 4) == 0)
    TRIG_WF = 1;
  else if (strncmp(request_command, "win", 3) == 0)
    WIN_WF = 1;
  else {
   fprintf(stderr, "Error: stp_req_waveform expects either a trig or win command.\n");
    return -1;
  }

  // send the command to the server
  n = send_stp_request(request_command);
  return n;
}


/* get sac waveform data from the server
 * return value = 0 end of requests
 *              >  0 # of data points delivered 
 *              = -1 error, chead contains the error
 * maxdata is the maximum number of data points wanted
 */
int stp_get_sac_waveform (char *chead, struct sac_header *shead, float **fdata, int maxdata) {

  int ret_val = 0, ndata, stp_val = 0;
  char temp[MAXLINE];
  float *data = NULL;
  char *buf = NULL;


  // this part is different for triggered and continuous data
  // because trig gets a filename twice and has a directory
  // trig might also have a *.evnt file
  if (TRIG_WF == 1) {

    if(GOT_DIR == 0) {
      stp_val = stp_dir_read(temp);
      if (stp_val != 1) {
	strcpy(chead, temp);
	if (stp_val == 0) {
	  close_waveserver();
	  TRIG_WF=0;
	  GOT_DIR=0;
	}
	return stp_val;
      }
      
      GOT_DIR = 1;
    }

    stp_val = stp_file_read(temp);
    if (stp_val != 1) {
      strcpy(chead, temp);
      if (stp_val == 0) {
	close_waveserver();
	TRIG_WF=0;
	GOT_DIR=0;
      }
      return stp_val;
    }
    
    
    // if we are getting the first file, go to the second
    if (strstr(temp, "evnt") != NULL) {
      ndata = stp_data_prep();
      stp_data_read(temp);
          
      stp_val = stp_file_read(temp);
      if (stp_val != 1) {
	strcpy(chead, temp);
	if (stp_val == 0) {
	  close_waveserver();
	  TRIG_WF=0;
	  GOT_DIR=0;
	}
	return stp_val;
      }
    
      stp_val = stp_file_read(temp);
      if (stp_val != 1) {
	strcpy(chead, temp);
	if (stp_val == 0) {
	  close_waveserver();
	  TRIG_WF=0;
	  GOT_DIR=0;
	}
	return stp_val;
      }
    
    }
    else {
      stp_val = stp_file_read(temp);
      if (stp_val != 1) {
	strcpy(chead, temp);
	if (stp_val == 0) {
	  close_waveserver();
	  TRIG_WF=0;
	  GOT_DIR=0;
	}
	return stp_val;
      }
    
    }
  }

  // for continous files, we only need to get the filename
  else if (WIN_WF == 1) {

    stp_val = stp_file_read(temp);
    if (stp_val != 1) {
      strcpy(chead, temp);
      WIN_WF=0;
      close_waveserver();
      return stp_val;
    }    
  }

  else {
    strcpy(chead, "Server is not expecting a request for waveforms.\n");
    return -1;
  }
  
  
  ndata = stp_data_prep();
  if (ndata == 0) {
    TRIG_WF = 0;
    WIN_WF = 0;
    return 0;
  }

  
  if ((ndata - 632) <= (maxdata * 4)) { 
    

    /* make sure data is the proper size */
    if(data != NULL)
      free(data);
    data = (float *) malloc(ndata);
    if (data == NULL) {
      strcpy(chead, "malloc failed.\n");
      TRIG_WF=0;
      GOT_DIR = 0;
      WIN_WF = 0;
      close_waveserver(); 
      return -1;
    }	
    
    ret_val = stp_sacinput_read(shead, data, ndata);
  }
  else {
    /* make sure data is the proper size */
    if(data != NULL)
      free(data);
    data = (float *) malloc(4*maxdata); 
    buf = (char *)malloc(ndata);
    if (data == NULL || buf == NULL) {
      strcpy(chead, "malloc failed.\n");
      TRIG_WF=0;
      GOT_DIR = 0;
      WIN_WF = 0;
      free(data);
      free(buf);
      close_waveserver(); 
      return -1;
    }	
    
    ret_val = stp_data_read(buf);
    buf_to_sac_input(buf, maxdata*4 + sizeof(struct sac_header), shead, data);
    ret_val = maxdata * 4 + sizeof(struct sac_header);
    free(buf);
    
  }

  if(ret_val <= 0) {
    close_waveserver();
    TRIG_WF = 0;
    GOT_DIR = 0;
    WIN_WF = 0;
  }
  *fdata = data;
  str2charsachdr(chead, shead);
  
  return ret_val;
}

/* converts a struct sac_header into a string */
int str2charsachdr(char *chead, struct sac_header *shead) {
  char date[64];
  struct tm tm;
  
  sprintf(date, "%d %d", shead->nzyear, shead->nzjday);
  
  strptime(date, "%Y %j", &tm);

  // for continous
  if (shead->evla != -12345) {
    sprintf(chead, " NBYTES: %d\n NET: %s\n STA: %s\n CHAN: %s\n DATETIME: %4d/%02d/%02d %02d:%02d:%02d:%d\n EVENT LOCATION: %.4f %.4f %4.2f\n", shead->npts, shead->knetwk, shead->kstnm, shead->kcmpnm, shead->nzyear, tm.tm_mon + 1, tm.tm_mday, shead->nzhour, shead->nzmin, shead->nzsec, shead->nzmsec, shead->evla, shead->evlo, shead->evdp);
  }
  // for events
  else {
    sprintf(chead, " NBYTES: %d\n NET: %s\n STA: %s\n CHAN: %s\n DATETIME: %4d/%02d/%02d %02d:%02d:%02d:%d\n STATION LOCATION: %.4f %.4f %.0f %.0f\n", shead->npts, shead->knetwk, shead->kstnm, shead->kcmpnm, shead->nzyear, tm.tm_mon + 1, tm.tm_mday, shead->nzhour, shead->nzmin, shead->nzsec, shead->nzmsec, shead->stla, shead->stlo, shead->stel, shead->stdp);
  }
  return 0;
}

/*****************************************************************************/
/*                             EVENT                                         */
/*****************************************************************************/

int stp_req_event(char *request_command) {
  int n;
  
  // make sure we have no data waiting
  // to prevent nested requests
  if (WIN_WF==1 || TRIG_WF==1 || EVENT==1 || ALTLOC==1 || PHASE==1 || STATION==1) {
    fprintf(stderr, "Event Request Error: another request is in progress.\n");
    return -1;
  }
  
  open_waveserver(server);

  
  // send defaults 
  send_defaults();
  free_defaults();
  
  // check that we are sending an appropriate command
  if (strncmp(request_command, "event", 5) == 0)
    EVENT = 1;
  else
    {
      fprintf(stderr, "Error: stp_req_event expects you to request an event.\n");
      return -1;
    }
  
  // send the event request
  n = send_stp_request(request_command);
  return n;
}

/* get events in string and struct format 
 * if stpevent is NULL, not struct format will be returned
 * but event_ptr must not be null
 * returns the number of event_ptr / stpevent returned
 * if an error occurs, returns neg. number of event_ptr / stpevent
 * (a return value of 0 probably indicates an error)
 */
int stp_get_event(char ***event_ptr, STPEVENT ***stpevent) {
  int n=0;
  char **e, **events;
  STPEVENT **event_struct, **es;

  if (event_ptr == NULL) {
    fprintf(stderr, "Error stp_get_event: event_ptr may not be NULL\n");
    return -n;
  }

  // allocate space for the array of event strings
  events = (char **) malloc(sizeof(char *)*(n+1));
  if (events == NULL) {
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // allocate space for the first event string
  events[0] = (char *) malloc(MAXLINE);
  if (events[0] == NULL) {
    free(events);
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // if an event has not been requested, it is an error
  if (EVENT != 1) {
    strcpy(events[n], "ERROR: Server is not expecting a request for events.\n");
    *event_ptr = events;
    return -n;
  }

  // allocate space for an array of STPEVENTS
  event_struct = (STPEVENT **) malloc(sizeof(STPEVENT *)*(n+1));
  if (event_struct == NULL) {
    strcpy(events[n], "ERROR: allocation of event_struct failed.\n");
    *event_ptr = events;
    return -n;
  }

  // read events into the string array in a loop
  while(stp_string_read(events[n]) > 0) {  

    // check for strings that indicate the end of the data
    // we shouldn't try to make these strings into a struct
    if (strstr(events[n], "No data") == 0 && strstr(events[n], "# Number") == 0) {
      // allocate space for a new STPEVENT
      event_struct[n] = (STPEVENT *) malloc(sizeof(STPEVENT));
      if (event_struct[n] == NULL) {
	strcpy(events[n], "ERROR: allocation of event_struct[n] failed.\n");
	*event_ptr = events;
	if (stpevent != NULL)
	  *stpevent = event_struct;
	return -n;
      }
      
      // create a new STPEVENT from the string
      evtstr2struct(events[n], event_struct[n]);
    }
    // no struct if not a valid string
    else
      event_struct[n] = NULL;

    n++;
    
    // reallocate the event array so that there is room for the next event
    e = realloc(events, sizeof(char *)*(n+1));
    if (e == NULL) {
      strcpy(events[n-1], "ERROR: reallocation of events failed.\n");
      *event_ptr = events;
      if (stpevent != NULL)
	*stpevent = event_struct;
      return -(n-1);
    }
    events = e;
    e = NULL;
    
    // allocate a string to put the next event into
    events[n] = (char *) malloc(MAXLINE);
    if (events[n] == NULL) {
       strcpy(events[n-1], "ERROR: allocation of event_struct failed.\n");
       *event_ptr = events;
       if (stpevent != NULL)
	 *stpevent = event_struct;
       return -(n-1);
    }

    // reallocate the event struct so that there is room for the next event
    es = realloc(event_struct, sizeof(STPEVENT *)*(n+1));
    if (es == NULL) {
      strcpy(events[n], "ERROR: allocation of event_struct failed.\n");
      *event_ptr = events;
      if (stpevent != NULL)
	*stpevent = event_struct;
      return -n;
    }
    event_struct = es;
    es = NULL;
  }
  // the last event_struct ends up being NULL because we did not read
  // a valid event on the last pull
  event_struct[n]=NULL;
  
  // make sure the information we've gotten gets passed 
   *event_ptr = events;
   if (stpevent != NULL)
     *stpevent = event_struct;

   EVENT = 0;
   close_waveserver();
   return n;
}

/* fills an event struct with the values found in a given
 * event string
 */
int evtstr2struct (char *str, STPEVENT *event){
  char lat[20], lon[20], depth[20], mag[20], qual[20];
  int i;

  // scan the event into variables
  sscanf(str, "%d %s %s %s %s %s %s %c %s", &event->eventid, &event->type, &event->datetime, lat, lon, depth, mag, &event->magtype, qual);
  
  // transfer string varibles to double
  event->lat = atof(lat);
  event->lon = atof(lon);
  event->depth = atof(depth);
  event->mag = atof(mag);
  event->qual = atof(qual);

  return 0;
}


/*****************************************************************************/
/*                             ALTLOC                                         */
/*****************************************************************************/

int stp_req_altloc(char *request_command) {
  int n;
  
  // make sure we have no data waiting
  // to prevent nested requests
  if (WIN_WF==1 || TRIG_WF==1 || EVENT==1 || ALTLOC==1 || PHASE==1 || STATION==1) {
    fprintf(stderr, "Altloc Request Error: another request is in progress.\n");
    return -1;
  }
  
  open_waveserver(server);

  
  // send defaults 
  send_defaults();
  free_defaults();
  
  // check that we are sending an appropriate command
  if (strncmp(request_command, "altloc", 6) == 0)
    ALTLOC = 1;
  else
    {
      fprintf(stderr, "Error: stp_req_altloc expects you to request an altloc.\n");
      return -1;
    }

  if (strstr(request_command, " -s ") != NULL) 
    DO_SHORT = 1;
  else if(strstr(request_command, " -l ") != NULL) 
    DO_LONG = 1;
  else
    DO_SHORT = 1;
  
  
  // send the event request
  n = send_stp_request(request_command);
  return n;
}

/* get altlocss in string and struct format 
 * if stpaltloc is NULL, not struct format will be returned
 * but altloc_ptr must not be null
 * returns the number of altloc_ptr / stpevent returned
 * if an error occurs, returns neg. number of altloc_ptr / stpaltloc
 * (a return value of 0 probably indicates an error)
 */
int stp_get_altloc(char ***altloc_ptr, STPALTLOC ***stpaltloc) {
  int n=0;
  char **e, **altlocs;
  STPALTLOC **altloc_struct, **es;

  if (altloc_ptr == NULL) {
    fprintf(stderr, "Error stp_get_altloc: altloc_ptr may not be NULL\n");
    return -n;
  }

  // allocate space for the array of altloc strings
  altlocs = (char **) malloc(sizeof(char *)*(n+1));
  if (altlocs == NULL) {
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // allocate space for the first altloc string
  altlocs[0] = (char *) malloc(MAXLINE);
  if (altlocs[0] == NULL) {
    free(altlocs);
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // if an altloc has not been requested, it is an error
  if (ALTLOC != 1) {
    strcpy(altlocs[n], "ERROR: Server is not expecting a request for altlocs.\n");
    *altloc_ptr = altlocs;
    return -n;
  }

  // allocate space for an array of STPALTLOCS
  altloc_struct = (STPALTLOC **) malloc(sizeof(STPALTLOC *)*(n+1));
  if (altloc_struct == NULL) {
    strcpy(altlocs[n], "ERROR: allocation of altloc_struct failed.\n");
    *altloc_ptr = altlocs;
    return -n;
  }

  // read altlocs into the string array in a loop
  while(stp_string_read(altlocs[n]) > 0) {  

    // check for strings that indicate the end of the data
    // we shouldn't try to make these strings into a struct
    if (strstr(altlocs[n], "No data") == 0 && strstr(altlocs[n], "# Number") == 0) {
      // allocate space for a new STPALTLOC
      altloc_struct[n] = (STPALTLOC *) malloc(sizeof(STPALTLOC));
      if (altloc_struct[n] == NULL) {
	strcpy(altlocs[n], "ERROR: allocation of altloc_struct[n] failed.\n");
	*altloc_ptr = altlocs;
	if (stpaltloc != NULL)
	  *stpaltloc = altloc_struct;
	return -n;
      }
      
      // create a new STPALTLOC from the string
      alocstr2struct(altlocs[n], altloc_struct[n]);
    }
    // no struct if not a valid string
    else
      altloc_struct[n] = NULL;

    n++;
    
    // reallocate the altloc array so that there is room for the next altloc
    e = realloc(altlocs, sizeof(char *)*(n+1));
    if (e == NULL) {
      strcpy(altlocs[n-1], "ERROR: reallocation of altlocs failed.\n");
      *altloc_ptr = altlocs;
      if (stpaltloc != NULL)
	*stpaltloc = altloc_struct;
      return -(n-1);
    }
    altlocs = e;
    e = NULL;
    
    // allocate a string to put the next altloc into
    altlocs[n] = (char *) malloc(MAXLINE);
    if (altlocs[n] == NULL) {
       strcpy(altlocs[n-1], "ERROR: allocation of altloc_struct failed.\n");
       *altloc_ptr = altlocs;
       if (stpaltloc != NULL)
	 *stpaltloc = altloc_struct;
       return -(n-1);
    }

    // reallocate the altloc struct so that there is room for the next altloc
    es = realloc(altloc_struct, sizeof(STPALTLOC *)*(n+1));
    if (es == NULL) {
      strcpy(altlocs[n], "ERROR: allocation of altloc_struct failed.\n");
      *altloc_ptr = altlocs;
      if (stpaltloc != NULL)
	*stpaltloc = altloc_struct;
      return -n;
    }
    altloc_struct = es;
    es = NULL; 
  }
  // the last altloc_struct ends up being NULL because we did not read
  // a valid altloc on the last pull
  altloc_struct[n]=NULL;
  
  // make sure the information we've gotten gets passed 
   *altloc_ptr = altlocs;
   if (stpaltloc != NULL)
     *stpaltloc = altloc_struct;

   ALTLOC = 0;
   DO_SHORT = 0;
   DO_LONG = 0;
   close_waveserver();
   return n;
}

/* fills an altloc struct with the values found in a given
 * altloc string
 */
int alocstr2struct (char *str, STPALTLOC *altloc){
  char lat[20], lon[20], depth[20], mag[20];
  char err_ns[20], err_ew[20], err_z[20];
  int i;

  // scan the altloc into variables
  /* eventid datetime lat long depth mag magtype source err_ns err_ew err_z
     num_primary num_secondary total_ps rms */

  if (DO_SHORT == 1) {
    sscanf(str, "%d %s %s %s %s %s %c %s", &altloc->eventid, &altloc->datetime, lat, lon, depth, mag, &altloc->magtype, &altloc->source);
  
    // transfer string varibles to double
    altloc->lat = atof(lat);
    altloc->lon = atof(lon);
    altloc->depth = atof(depth);
    altloc->mag = atof(mag);
    altloc->err_ns = 0;
    altloc->err_ew = 0;
    altloc->err_z = 0;
    altloc->nop = 0;
    altloc->nos = 0;
    altloc->totps = 0;
    altloc->rms = 0;

  }

  if (DO_LONG == 1) {
    sscanf(str, "%d %s %s %s %s %s %c %s %s %s %s %d %d %d %s", &altloc->eventid, &altloc->datetime, lat, lon, depth, mag, &altloc->magtype, &altloc->source, err_ns, err_ew, err_z, &altloc->nop, &altloc->nos, &altloc->totps, &altloc->rms);
  
    // transfer string varibles to double
    altloc->lat = atof(lat);
    altloc->lon = atof(lon);
    altloc->depth = atof(depth);
    altloc->mag = atof(mag);
    altloc->err_ns = atof(err_ns);
    altloc->err_ew = atof(err_ew);
    altloc->err_z = atof(err_z);  
  }

  return 0;
}


/*****************************************************************************/
/*                         STATION                                           */
/*****************************************************************************/

/* made so that it always has a "-l" argument to make the 
 * retrieval of data simpler (only one form as oposed to two)
 */
int stp_req_station(char *request_command) {
  char temp[MAXLINE];
  int n, i, minl;

  
  // make sure we have no data waiting
  // to prevent nested requests
  if (WIN_WF==1 || TRIG_WF==1 || EVENT==1 || ALTLOC==1 ||  PHASE==1 || STATION==1) {
    fprintf(stderr, "another request in progress\n");
    return -1;
  }

  // open the waveserver
  open_waveserver(server);
  
  // send defaults 
  send_defaults();
  free_defaults();
  
  // check that we are sending an appropriate command
  if (strncmp(request_command, "sta", 3) == 0)
    STATION = 1;
  else
    return -1;

  // add -l argument if it does not exist
  minl = 0;
  for (i = 0; i < strlen(request_command) - 1; i++) {
    if (request_command[i] == '-' && request_command[i+1] == 'l') {
      minl = 1;
      break;
    }
  }
  if (minl == 0) {
    strcpy(temp, request_command);
    temp[4] = '-';
    temp[5] = 'l';
    temp[6] = ' ';
    for (i = 7; i < strlen(request_command) + 3; i++) {
      temp[i] = request_command[i-3];
    }
    temp[i] = '\0';
  }
  else
    strcpy(temp, request_command);

  // send the request
  n = send_stp_request(temp);
  return n;
}

/* get station in string and struct format
 * if stpstation is NULL, it will not be used
 * but station_ptr cannot be NULL
 * returns number of station_ptrs / stpstations returned
 * if there is an error, returns the negative of this (a return value of 
 * zero probably indicates an error).
 */
int stp_get_station(char ***station_ptr, STPSTA ***stpstation) {
  int n=0;
  char **e, **stations;
  STPSTA **sta_struct, **es;


  // allocate space for the array of stations
  stations = (char **) malloc(sizeof(char *)*(n+1));
  if (stations == NULL) {
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // allocate space for the first station
  stations[n] = (char *) malloc(MAXLINE);
  if (stations[n] == NULL) {
    free(stations);
    fprintf(stderr, "Allocation failed\n");
    return -n;
  }

  // check whether the server has station information
  if (STATION != 1) {
    strcpy(stations[n], "ERROR: Server is not expecting a request for stations.\n");
    *station_ptr = stations;
    return -n;
  }

  // allocate space for an array of STPSTA
  sta_struct = (STPSTA **) malloc(sizeof(STPSTA *)*(n+1));
  if (sta_struct == NULL) {
    strcpy(stations[n], "Error: Allocation failed for station struct.\n");
    *station_ptr = stations;
    return -n;
  }

  // read int stations in a loop
  while(stp_string_read(stations[n]) > 0) {  

    // check for information strings, these shouldn't be made into structs
    if (strstr(stations[n], "No data") == 0 && strstr(stations[n], "# of stations") == 0) {
      sta_struct[n] = (STPSTA *) malloc(sizeof(STPSTA));
      if (sta_struct[n] == NULL) {	
	strcpy(stations[n], "Error: Allocation failed for station struct[n].\n");
	*station_ptr = stations;
	if (stpstation != NULL) 
	  *stpstation = sta_struct;
	return -n;
      }
      
      // make the string into a struct
      stastr2struct(stations[n], sta_struct[n]);

    }
    // no struct
    else
      sta_struct[n] = NULL;

    n++;
    
    // allocate more space for the array of strings
    e = realloc(stations, sizeof(char *)*(n+1));
    if (e == NULL) {
      strcpy(stations[n-1], "Error: Reallocation failed for stations.\n");
      *station_ptr = stations;
      if (stpstation != NULL) 
	*stpstation = sta_struct;
      return -(n-1);
    }
    stations = e;
    e = NULL;

    // allocate a new string
    stations[n] = (char *) malloc(MAXLINE);
    if (stations[n] == NULL) {
       strcpy(stations[n-1], "Error: Reallocation failed for stations[n].\n");
       *station_ptr = stations;
       if (stpstation != NULL) 
	 *stpstation = sta_struct;
       return -(n-1);
    }

    // allocate more space for the array of structs
    es = realloc(sta_struct, sizeof(STPSTA *)*(n+1));
    if (es == NULL) {
      strcpy(stations[n-1], "Error: Reallocation failed for sta_struct.\n");
      *station_ptr = stations;
      if (stpstation != NULL) 
	*stpstation = sta_struct;
      return -n;
    }
    sta_struct = es;
    es = NULL;
    
  }
  // the last struct is always null because the loop stopped
  sta_struct[n]=NULL;
  
  // make sure these values are passed back
  *station_ptr = stations;

  if (stpstation != NULL) 
    *stpstation = sta_struct;

  STATION = 0;
  close_waveserver();
  return n;
}

/* fills a station struct with the values found in a given
 * station string
 */
int stastr2struct (char *str, STPSTA *station){
  char temp[20], lat[20], lon[20], elev[20]; 
  int i, j;

  sscanf(str, "%s %s %s %s\n", &temp, lat, lon, elev);
  
  for (i=0; temp[i]!='.' && i<4; i++) {
    station->net[i] = temp[i];
  }
  station->net[i] = '\0';
  i++;
  for (j=0; temp[i]!='\0' && j<4; i++) {
    station->sta[j] = temp[i];
    j++;
  }
  station->sta[j]='\0';
  
  station->lat = atof(lat);
  station->lon = atof(lon);
  station->elev = atof(elev);

  return 0;
}

/*****************************************************************************/
/*                              PHASE                                        */
/*****************************************************************************/


/* function to request phase info from the stp server */
int stp_req_phase(char *request_command) {
  int n;
  
  // make sure we have no data waiting
  // to prevent nested requests
  if (WIN_WF==1 || TRIG_WF==1 || EVENT==1 || ALTLOC==1 || PHASE==1 || STATION==1) {
    fprintf(stderr, "another request in progress\n");
    return -1;
  }
  
  open_waveserver(server);
  
  // send defaults 
  send_defaults();
  free_defaults();
  
  // check that we are sending an appropriate command
  if (strncmp(request_command, "phase", 5) == 0)
    PHASE = 1;
  else
    return -1;
 
   n = send_stp_request(request_command);
  
  return n;
}


/* get the phase info in two structs, a struct of strings and a struct
 * of structs. NULL can be given for strphase or stpphase and no data
 * will be returned in that form.
 * returns the number of events, or the negative of that for an error 
 * (a return value of 0 may indicate an error)
 */
int stp_get_phase(char **phase, STRINGPHASE ***strphase, STPPHASE ***stpphase) {

  int evtnum = 0; 
  int index, i;

  // STRINGPHASE variables
  STRINGPHASE **strp = NULL, **s;
  STRINGPHASE *temp;
  char **chray;

  // STPPHASE variables
  STPPHASE **stpp = NULL, **p;
  STPPHASE *stptemp;
  PHAINFO **pi;
  
  // this variable must not be null
  if (phase == NULL) {
    fprintf (stderr, "first variable should not be NULL\n");
    return -evtnum;
  }
  // allocate space for error messages and lines from the server
  *phase = (char *) malloc(MAXLINE);
  if (*phase == NULL) {
    fprintf(stderr, "Allocation failed.\n");
    return -evtnum;
  }

  // make sure that the server has phases
  if (PHASE != 1) {
    strcpy(*phase, "ERROR: Server is not expecting a request for phases.\n");
    return -evtnum;
  }

  // read from the server in a loop
  while (stp_string_read(phase[0]) > 0) {

    // new event to make space for
    if (atoi(*phase) != 0) {
      
      evtnum++;
      
      /********************** STRINGPHASE*************************/
      if (strphase != NULL) {
	
	s = realloc(strp, evtnum * sizeof(STRINGPHASE *));
	if (s == NULL) {
	  
	  strcpy(*phase, "Reallocation failed for strp\n");
	    if (strphase != NULL) 
	      *strphase = strp;
	    if (stpphase != NULL)
	      *stpphase = stpp;
	  return -evtnum;
	}
	strp = s;
	s = NULL;

	strp[evtnum - 1] = (STRINGPHASE *)malloc(sizeof(STRINGPHASE));
	if (strp[evtnum - 1] == NULL) {
	  strcpy(*phase, "Allocation failed for strp[n]\n");
	  if (strphase != NULL) 
	    *strphase = strp;
	  if (stpphase != NULL)
	    *stpphase = stpp;
	  return -evtnum;
	}
	
	// use the temp variable to avoid really long variable names    
	temp = strp[evtnum - 1];
	temp->evtinfo = malloc(MAXLINE);
	temp->numphastr = 0;
	temp->phastr = NULL;
	strcpy(temp->evtinfo, phase[0]);
      }
      
      /*********************** STPPHASE **************************/
      if (stpphase != NULL) {
		
	p = realloc(stpp, evtnum * sizeof(STPPHASE *));
	
	if (p == NULL) {
	  strcpy(*phase, "Reallocation failed for stpp\n");
	  if (strphase != NULL) 
	    *strphase = strp;
	  if (stpphase != NULL)
	    *stpphase = stpp;
	  return -evtnum;
	}
	stpp = p;
	p = NULL;
	
	stpp[evtnum - 1] = (STPPHASE *) malloc(sizeof(STPPHASE));
	if (stpp[evtnum - 1] == NULL) {
	  strcpy(*phase, "Allocation failed for stpp[n]\n");
	  if (strphase != NULL) 
	    *strphase = strp;
	  if (stpphase != NULL)
	    *stpphase = stpp;
	  return -evtnum;
	}
	
	// use stptemp to avoid really long variable names
	stptemp = stpp[evtnum - 1];
	stptemp->evtinfo = malloc(sizeof(STPEVENT));
	stptemp->numphainfo = 0;
	stptemp->phainfo = NULL;
	evtstr2struct(phase[0], stptemp->evtinfo);
      }
    } 
      // getting phase info for the last event 
      else {  
	
      /********************* STRINGPHASE **************************/
	if (strphase != NULL) {
	  index = temp->numphastr;
	  temp->numphastr++;      
	  
	  chray = realloc(temp->phastr, temp->numphastr * sizeof(char *));
	  if (chray == NULL) {
	    strcpy(*phase, "Reallocation failed for temp->phastr\n");
	    if (strphase != NULL) 
	      *strphase = strp;
	    if (stpphase != NULL)
	      *stpphase = stpp;
	    return -evtnum;
	  }
	  temp->phastr = chray;
	  chray = NULL;
	  
	  temp->phastr[index] = malloc(MAXLINE);
	  if (temp->phastr[index] == NULL) {
	    strcpy(*phase, "Allocation failed for temp->phastr[n]\n");
	    if (strphase != NULL) 
	      *strphase = strp;
	    if (stpphase != NULL)
	      *stpphase = stpp;
	    return -evtnum;
	  }
	  
	  strcpy(temp->phastr[index], phase[0]);
	}
	
	/****************** STPPHASE *******************************/
	
	if (stpphase != NULL) {
	  index = stptemp->numphainfo;
	  stptemp->numphainfo++;
	  
	  pi = realloc(stptemp->phainfo, stptemp->numphainfo * sizeof(PHAINFO *));
	  if (pi == NULL) {
	    strcpy(*phase, "Reallocation failed for stptemp->phainfo\n");
	    if (strphase != NULL) 
	      *strphase = strp;
	    if (stpphase != NULL)
	      *stpphase = stpp;
	    return -evtnum;
	  }
	  stptemp->phainfo = pi;
	  pi = NULL;
	  
	  stptemp->phainfo[index] = malloc(sizeof(PHAINFO));
	  if (stptemp->phainfo[index] == NULL) {
	    strcpy(*phase, "Allocation failed for stptemp->phainfo[n]\n");
	    if (strphase != NULL) 
	      *strphase = strp;
	    if (stpphase != NULL)
	      *stpphase = stpp;
	    return -evtnum;
	  }
	  
	  phastr2struct(phase[0], stptemp->phainfo[index]);
	}
      }
    
  } // end while
  

  // make sure the data gets passed
  if (strphase != NULL) 
    *strphase = strp;
  if (stpphase != NULL)
    *stpphase = stpp;

  PHASE = 0;
  close_waveserver();
  return evtnum;
}

int phastr2struct (char *str, PHAINFO *pha) {

  /* float strings */
  char lat[20], lon[20], elev[20];
  char pickqual[20], epidist[20], time[20];
  double temp;

  // read in from the string
  sscanf(str, "%s %s %s %s %s %s %c %c%c %c %s %s %s", &pha->net, &pha->sta, &pha->chan, lat, lon, elev, &pha->phase, &pha->motion1[0], &pha->motion1[1], &pha->soqual, pickqual, epidist, time);
  pha->motion1[2] = '\0';

  // convert strings to double
  pha->lat = atof(lat); 
  pha->lon = atof(lon);
  pha->elev = atof(elev);
  pha->pickqual = atof(pickqual);
  pha->epidist = atof(epidist);
  pha->time = atof(time);
  
  return 0;

} 
